@@ -1,19 +1,43 @@
-from datetime import datetime
+from __future__ import annotations
+
+from datetime import datetime, timedelta
+from functools import lru_cache
 import os
 import os.path as op
+from pathlib import Path
 import re
+import sys
+from typing import (
+    TYPE_CHECKING,
+    Any,
+    Callable,
+    Dict,
+    Iterable,
+    List,
+    Optional,
+    Tuple,
+    Type,
+    TypeVar,
+    Union,
+    cast,
+)
 from uuid import uuid4
+from xml.dom.minidom import parseString
 
 from dandischema import models
+import requests
+import tenacity
 
 from . import __version__, get_logger
 from .dandiset import Dandiset
+from .misctypes import Digest
 from .pynwb_utils import (
     _get_pynwb_metadata,
     get_neurodata_types,
     get_nwb_version,
     ignore_benign_pynwb_warnings,
     metadata_cache,
+    nwb_has_external_links,
 )
 from .utils import ensure_datetime, get_mime_type, get_utcnow_datetime
 
@@ -21,7 +45,7 @@ lgr = get_logger()
 
 
 @metadata_cache.memoize_path
-def get_metadata(path):
+def get_metadata(path: Union[str, Path]) -> Optional[dict]:
     """Get selected metadata from a .nwb file or a dandiset directory
 
     If a directory given and it is not a Dandiset, None is returned
@@ -42,11 +66,16 @@ def get_metadata(path):
     if op.isdir(path):
         try:
             dandiset = Dandiset(path)
-            return dandiset.metadata
+            return cast(dict, dandiset.metadata)
         except ValueError as exc:
             lgr.debug("Failed to get metadata for %s: %s", path, exc)
             return None
 
+    if nwb_has_external_links(path):
+        raise NotImplementedError(
+            f"NWB files with external links are not supported: {path}"
+        )
+
     # First read out possibly available versions of specifications for NWB(:N)
     meta["nwb_version"] = get_nwb_version(path)
 
@@ -90,25 +119,22 @@ def get_metadata(path):
     return meta
 
 
-def _parse_iso8601(age):
+def _parse_iso8601(age: str) -> List[str]:
     """checking if age is proper iso8601, additional formatting"""
     # allowing for comma instead of ., e.g. P1,5D
     age = age.replace(",", ".")
     pattern = (
-        "^P(?!$)(\\d+(?:\\.\\d+)?Y)?(\\d+(?:\\.\\d+)?M)?(\\d+(?:\\.\\d+)?W)?(\\d+(?:\\.\\d+)?D)?"
-        "(T(?=\\d)(\\d+(?:\\.\\d+)?H)?(\\d+(?:\\.\\d+)?M)?(\\d+(?:\\.\\d+)?S)?)?$"
+        r"^P(?!$)(\d+(?:\.\d+)?Y)?(\d+(?:\.\d+)?M)?(\d+(?:\.\d+)?W)?(\d+(?:\.\d+)?D)?"
+        r"(T(?=\d)(\d+(?:\.\d+)?H)?(\d+(?:\.\d+)?M)?(\d+(?:\.\d+)?S)?)?$"
     )
-
-    matchstr = re.match(pattern, age, flags=re.I)
-    if matchstr:
-        age_frm = [matchstr.group(i) for i in range(1, 6) if matchstr.group(i)]
-        age_frm = ["P"] + age_frm
-        return age_frm
+    m = re.match(pattern, age, flags=re.I)
+    if m:
+        return ["P"] + [m[i] for i in range(1, 6) if m[i]]
     else:
-        raise ValueError(f"ISO 8601 expected, but {age} was received")
+        raise ValueError(f"ISO 8601 expected, but {age!r} was received")
 
 
-def _parse_age_re(age, unit, tp="date"):
+def _parse_age_re(age: str, unit: str, tp: str = "date") -> Tuple[str, Optional[str]]:
     """finding parts that have <value> <unit> in various forms"""
 
     if unit == "Y":
@@ -125,65 +151,57 @@ def _parse_age_re(age, unit, tp="date"):
         pat_un = "(min|m(inute)?)"
     elif unit == "S":
         pat_un = "(sec|s(econd)?)"
-
-    pattern = rf"(\d+\.?\d*)\s*({pat_un}s?)"
-    matchstr = re.match(pattern, age, flags=re.I)
-    swap_flag = False
-    if matchstr is None:
-        # checking pattern with "unit" word
-        pattern_unit = rf"(\d+\.?\d*)\s*units?:?\s*({pat_un}s?)"
-        matchstr = re.match(pattern_unit, age, flags=re.I)
-        if matchstr is None:
-            # checking patter with swapped order
-            pattern = rf"({pat_un}s?)\s*(\d+\.?\d*)"
-            matchstr = re.match(pattern, age, flags=re.I)
-            swap_flag = True
-            if matchstr is None:
-                return age, None
-    if swap_flag:
-        qty = matchstr.group(3)
     else:
-        qty = matchstr.group(1)
+        raise AssertionError(f"Unexpected unit: {unit!r}")
+
+    m = re.match(rf"(\d+\.?\d*)\s*({pat_un}s?)", age, flags=re.I)
+    swap_flag = False
+    if m is None:
+        # checking pattern with "unit" word
+        m = re.match(rf"(\d+\.?\d*)\s*units?:?\s*({pat_un}s?)", age, flags=re.I)
+        if m is None:
+            # checking pattern with swapped order
+            m = re.match(rf"({pat_un}s?)\s*(\d+\.?\d*)", age, flags=re.I)
+            swap_flag = True
+            if m is None:
+                return age, None
+    qty = m[3 if swap_flag else 1]
     if "." in qty:
         qty = float(qty)
         if int(qty) == qty:
             qty = int(qty)
     else:
         qty = int(qty)
-    age_rem = age.replace(matchstr.group(0), "")
-    age_rem = age_rem.strip()
-    return age_rem, f"{qty}{unit}"
+    return (age[: m.start()] + age[m.end() :]).strip(), f"{qty}{unit}"
 
 
-def _parse_hours_format(age):
+def _parse_hours_format(age: str) -> Tuple[str, List[str]]:
     """parsing format 0:30:10"""
-    pattern = r"\s*(\d\d?):(\d\d):(\d\d)"
-    matchstr = re.match(pattern, age, flags=re.I)
-    if matchstr:
-        time_part = f"T{int(matchstr.group(1))}H{int(matchstr.group(2))}M{int(matchstr.group(3))}S"
-        age_rem = age.replace(matchstr.group(0), "")
-        age_rem = age_rem.strip()
-        return age_rem, [time_part]
+    m = re.match(r"\s*(\d\d?):(\d\d):(\d\d)", age)
+    if m:
+        time_part = f"T{int(m[1])}H{int(m[2])}M{int(m[3])}S"
+        return (age[: m.start()] + age[m.end() :]).strip(), [time_part]
     else:
         return age, []
 
 
-def _check_decimal_parts(age_parts):
+def _check_decimal_parts(age_parts: List[str]) -> bool:
     """checking if decimal parts are only in the lowest order component"""
     # if the last part is the T component I have to separate the parts
     if "T" in age_parts[-1]:
-        pattern_time = "^T(\\d+(?:\\.\\d+)?H)?(\\d+(?:\\.\\d+)?M)?(\\d+(?:\\.\\d+)?S)?"
-        matchstr = re.match(pattern_time, age_parts[-1], flags=re.I)
-        time_parts = [matchstr.group(i) for i in range(1, 3) if matchstr.group(i)]
-        age_parts = age_parts[:-1] + time_parts
+        m = re.match(
+            r"^T(\d+(?:\.\d+)?H)?(\d+(?:\.\d+)?M)?(\d+(?:\.\d+)?S)?",
+            age_parts[-1],
+            flags=re.I,
+        )
+        if m is None:
+            raise ValueError(f"Failed to parse the trailing part of age {age_parts[-1]!r}")
+        age_parts = age_parts[:-1] + [m[i] for i in range(1, 3) if m[i]]
     decim_part = ["." in el for el in age_parts]
-    if any(decim_part) and any(decim_part[:-1]):
-        return False
-    else:
-        return True
+    return not (any(decim_part) and any(decim_part[:-1]))
 
 
-def parse_age(age):
+def parse_age(age: Optional[str]) -> Tuple[str, str]:
     """
     Parsing age field and converting into an ISO 8601 duration
 
@@ -193,25 +211,23 @@ def parse_age(age):
 
     Returns
     -------
-    str
+    Tuple[str, str]
     """
 
     if not age:
-        raise ValueError("age is empty")
+        raise ValueError("Age is empty")
 
     age_orig = age
 
-    if "gestation" in age.lower():
-        pattern_time = "^gest[a-z]*"
-        matchstr = re.match(pattern_time, age, flags=re.I)
-        age = age.replace(matchstr.group(0), "")
+    if age.lower().startswith("gestation"):
+        age = re.sub("^gest[a-z]*", "", age, flags=re.I)
         ref = "Gestational"
     else:
         ref = "Birth"
 
     age = age.strip()
 
-    if age[0] == "P":
+    if age.startswith("P"):
         age_f = _parse_iso8601(age)
     else:  # trying to figure out any free form
         # removing some symbols
@@ -219,9 +235,9 @@ def parse_age(age):
             age = age.replace(symb, " ")
         age = age.strip()
         if not age:
-            raise ValueError("age doesn't have any information")
+            raise ValueError("Age doesn't have any information")
 
-        date_f = []
+        date_f: List[str] = []
         for unit in ["Y", "M", "W", "D"]:
             if not age:
                 break
@@ -232,7 +248,7 @@ def parse_age(age):
                 date_f = ["P", part_f]
 
         if ref == "Birth":
-            time_f = []
+            time_f: List[str] = []
             for un in ["H", "M", "S"]:
                 if not age:
                     break
@@ -251,19 +267,19 @@ def parse_age(age):
             age_f = date_f
         if set(age) - {" ", ".", ",", ":", ";"}:
             raise ValueError(
-                f"not able to parse the age: {age_orig}, no rules to convert: {age}"
+                f"Cannot parse age {age_orig!r}: no rules to convert {age!r}"
             )
 
     # checking if there are decimal parts in the higher order components
     if not _check_decimal_parts(age_f):
         raise ValueError(
-            f"decimal fraction allowed in the lowest order part only,"
-            f" but {age} was received "
+            f"Decimal fraction allowed in the lowest order part only,"
+            f" but {age!r} was received"
         )
     return "".join(age_f), ref
 
 
-def extract_age(metadata):
+def extract_age(metadata: dict) -> Optional[models.PropertyValue]:
     try:
         dob = ensure_datetime(metadata["date_of_birth"])
         start = ensure_datetime(metadata["session_start_time"])
@@ -271,7 +287,7 @@ def extract_age(metadata):
         if metadata.get("age") is not None:
             duration, ref = parse_age(metadata["age"])
         else:
-            return ...
+            return None
     else:
         duration, ref = timedelta2duration(start - dob), "Birth"
     return models.PropertyValue(
@@ -283,7 +299,7 @@ def extract_age(metadata):
     )
 
 
-def timedelta2duration(delta):
+def timedelta2duration(delta: timedelta) -> str:
     """
     Convert a datetime.timedelta to ISO 8601 duration format
 
@@ -299,7 +315,7 @@ def timedelta2duration(delta):
     if delta.days:
         s += f"{delta.days}D"
     if delta.seconds or delta.microseconds:
-        sec = delta.seconds
+        sec: Union[int, float] = delta.seconds
         if delta.microseconds:
             # Don't add when microseconds is 0, so that sec will be an int then
             sec += delta.microseconds / 1e6
@@ -309,7 +325,7 @@ def timedelta2duration(delta):
     return s
 
 
-def extract_sex(metadata):
+def extract_sex(metadata: dict) -> Optional[models.SexType]:
     value = metadata.get("sex", None)
     if value is not None and value != "":
         value = value.lower()
@@ -332,78 +348,178 @@ def extract_sex(metadata):
             raise ValueError(f"Cannot interpret sex field: {value}")
         return models.SexType(identifier=value_id, name=value)
     else:
-        return ...
+        return None
 
 
-def extract_species(metadata):
-    value = metadata.get("species", None)
-    if value is not None and value != "":
-        value = value.lower()
-        if "mouse" in value or value.startswith("mus"):
-            value_id = "http://purl.obolibrary.org/obo/NCBITaxon_10090"
-            value = "House mouse"
-        elif "human" in value or value.startswith("homo"):
-            value_id = "http://purl.obolibrary.org/obo/NCBITaxon_9606"
-            value = "Human"
-        elif "norvegicus" in value:
-            value_id = "http://purl.obolibrary.org/obo/NCBITaxon_10116"
-            value = "Rat; Norway rat; rats; Brown rat"
-        elif "rattus rattus" in value:
-            value_id = "http://purl.obolibrary.org/obo/NCBITaxon_10117"
-            value = "Black rat; Roof rat; House rat"
-        elif "rat" in value:
-            value_id = "http://purl.obolibrary.org/obo/NCBITaxon_10116"
-            value = "Rat; Norway rat; rats; Brown rat"
-        elif "mulatta" in value or "rhesus" in value:
-            value_id = "http://purl.obolibrary.org/obo/NCBITaxon_9544"
-            value = "Rhesus monkey"
-        elif "jacchus" in value:
-            value_id = "http://purl.obolibrary.org/obo/NCBITaxon_9483"
-            value = "Common marmoset"
-        elif "melanogaster" in value or "fruit fly" in value:
-            value_id = "http://purl.obolibrary.org/obo/NCBITaxon_7227"
-            value = "Common fruit fly"
-        elif value.startswith("http"):
-            value_id = value
-            value = None
+species_map = [
+    (
+        ["mouse"],
+        "mus",
+        "http://purl.obolibrary.org/obo/NCBITaxon_10090",
+        "Mus musculus - House mouse",
+    ),
+    (
+        ["human"],
+        "homo",
+        "http://purl.obolibrary.org/obo/NCBITaxon_9606",
+        "Homo sapiens - Human",
+    ),
+    (
+        ["norvegicus"],
+        None,
+        "http://purl.obolibrary.org/obo/NCBITaxon_10116",
+        "Rattus norvegicus - Norway rat",
+    ),
+    (
+        ["rattus rattus"],
+        None,
+        "http://purl.obolibrary.org/obo/NCBITaxon_10117",
+        "Rattus rattus - Black rat",
+    ),
+    (
+        ["rat"],
+        None,
+        "http://purl.obolibrary.org/obo/NCBITaxon_10116",
+        "Rattus norvegicus - Norway rat",
+    ),
+    (
+        ["mulatta", "rhesus"],
+        None,
+        "http://purl.obolibrary.org/obo/NCBITaxon_9544",
+        "Macaca mulatta - Rhesus monkey",
+    ),
+    (
+        ["jacchus"],
+        None,
+        "http://purl.obolibrary.org/obo/NCBITaxon_9483",
+        "Callithrix jacchus - Common marmoset",
+    ),
+    (
+        ["melanogaster", "fruit fly"],
+        None,
+        "http://purl.obolibrary.org/obo/NCBITaxon_7227",
+        "Drosophila melanogaster - Fruit fly",
+    ),
+]
+
+
+@lru_cache(maxsize=None)
+@tenacity.retry(
+    reraise=True,
+    stop=tenacity.stop_after_attempt(3),
+    wait=tenacity.wait_exponential(exp_base=1.25, multiplier=1.25),
+)
+def parse_purlobourl(
+    url: str, lookup: Optional[Tuple[str, ...]] = None
+) -> Optional[Dict[str, str]]:
+    """Parse an Ontobee URL to return properties of a Class node
+
+    :param url: Ontobee URL
+    :param lookup: list of XML nodes to lookup
+    :return: dictionary containing found nodes
+    """
+
+    req = requests.get(url, allow_redirects=True)
+    req.raise_for_status()
+    doc = parseString(req.text)
+    for elfound in doc.getElementsByTagName("Class"):
+        if (
+            "rdf:about" in elfound.attributes.keys()
+            and elfound.attributes.getNamedItem("rdf:about").value == url
+        ):
+            break
+    else:
+        return None
+    values = {}
+    if lookup is None:
+        lookup = ("rdfs:label", "oboInOwl:hasExactSynonym")
+    for key in lookup:
+        elchild = elfound.getElementsByTagName(key)
+        if elchild:
+            elchild = elchild[0]
+            values[key] = elchild.childNodes[0].nodeValue.capitalize()
+    return values
+
+
+def extract_species(metadata: dict) -> Optional[models.SpeciesType]:
+    value_orig = metadata.get("species", None)
+    value_id = None
+    if value_orig is not None and value_orig != "":
+        value = value_orig.lower().rstrip("/")
+        if value.startswith("http://purl.obolibrary.org/obo/NCBITaxon_".lower()):
+            for common_names, prefix, uri, name in species_map:
+                if value.split("//")[1] == uri.lower().rstrip("/").split("//")[1]:
+                    value_id = uri
+                    value = name
+                    break
+            if value_id is None:
+                value_id = value_orig
+                lookup = ("rdfs:label", "oboInOwl:hasExactSynonym")
+                try:
+                    result = parse_purlobourl(value_orig, lookup=lookup)
+                except ConnectionError:
+                    value = None
+                else:
+                    value = None
+                    if result is not None:
+                        value = " - ".join(
+                            [result[key] for key in lookup if key in result]
+                        )
         else:
+            for common_names, prefix, uri, name in species_map:
+                if any(key in value for key in common_names) or (
+                    prefix and value.startswith(prefix)
+                ):
+                    value_id = uri
+                    value = name
+                    break
+        if value_id is None:
             raise ValueError(
                 f"Cannot interpret species field: {value}. Please "
-                "contact help@dandiarchive.org to add your species."
+                "contact help@dandiarchive.org to add your species. "
+                "You can also put the entire url from NCBITaxon "
+                "(http://www.ontobee.org/ontology/NCBITaxon) into "
+                "your species field in your NWB file. For example: "
+                "http://purl.obolibrary.org/obo/NCBITaxon_9606 is the "
+                "url for the species Homo sapiens. Please note that "
+                "this url is case sensitive."
             )
-        return models.SpeciesType(
-            identifier=value_id, name=value.capitalize() if value is not None else None
-        )
+        return models.SpeciesType(identifier=value_id, name=value)
     else:
-        return ...
+        return None
 
 
-def extract_assay_type(metadata):
+def extract_assay_type(metadata: dict) -> Optional[List[models.AssayType]]:
     if "assayType" in metadata:
         return [models.AssayType(identifier="assayType", name=metadata["assayType"])]
     else:
-        return ...
+        return None
 
 
-def extract_anatomy(metadata):
+def extract_anatomy(metadata: dict) -> Optional[List[models.Anatomy]]:
     if "anatomy" in metadata:
         return [models.Anatomy(identifier="anatomy", name=metadata["anatomy"])]
     else:
-        return ...
+        return None
 
 
-def extract_model(modelcls, metadata, **kwargs):
-    m = modelcls.unvalidated()
+M = TypeVar("M", bound=models.DandiBaseModel)
+
+
+def extract_model(modelcls: Type[M], metadata: dict, **kwargs: Any) -> M:
+    m = cast(M, modelcls.unvalidated())
     for field in m.__fields__.keys():
         value = kwargs.get(field, extract_field(field, metadata))
-        if value is not Ellipsis:
+        if value is not None:
             setattr(m, field, value)
     # return modelcls(**m.dict())
     return m
 
 
-def extract_model_list(modelcls, id_field, id_source, **kwargs):
-    def func(metadata):
+def extract_model_list(
+    modelcls: Type[M], id_field: str, id_source: str, **kwargs: Any
+) -> Callable[[dict], List[M]]:
+    def func(metadata: dict) -> List[M]:
         m = extract_model(
             modelcls, metadata, **{id_field: metadata.get(id_source)}, **kwargs
         )
@@ -415,8 +531,8 @@ def extract_model_list(modelcls, id_field, id_source, **kwargs):
     return func
 
 
-def extract_wasDerivedFrom(metadata):
-    derived_from = None
+def extract_wasDerivedFrom(metadata: dict) -> Optional[List[models.BioSample]]:
+    derived_from: Optional[List[models.BioSample]] = None
     for field, sample_name in [
         ("tissue_sample_id", "tissuesample"),
         ("slice_id", "slice"),
@@ -434,18 +550,18 @@ def extract_wasDerivedFrom(metadata):
 
 
 extract_wasAttributedTo = extract_model_list(
-    models.Participant, "identifier", "subject_id", id=...
+    models.Participant, "identifier", "subject_id", id=None
 )
 
 
-def extract_session(metadata: dict) -> list:
+def extract_session(metadata: dict) -> Optional[List[models.Session]]:
     probe_ids = metadata.get("probe_ids", [])
     if isinstance(probe_ids, str):
         probe_ids = [probe_ids]
-    probes = []
-    for val in probe_ids:
-        probes.append(models.Equipment(identifier=f"probe:{val}", name="Ecephys Probe"))
-    probes = probes or None
+    probes = [
+        models.Equipment(identifier=f"probe:{val}", name="Ecephys Probe")
+        for val in probe_ids
+    ] or None
     session_id = None
     if metadata.get("session_id") is not None:
         session_id = str(metadata["session_id"])
@@ -462,14 +578,16 @@ def extract_session(metadata: dict) -> list:
     ]
 
 
-def extract_digest(metadata):
+def extract_digest(
+    metadata: dict,
+) -> Optional[Dict[models.DigestType, str]]:
     if "digest" in metadata:
         return {models.DigestType[metadata["digest_type"]]: metadata["digest"]}
     else:
-        return ...
+        return None
 
 
-FIELD_EXTRACTORS = {
+FIELD_EXTRACTORS: Dict[str, Callable[[dict], Any]] = {
     "wasDerivedFrom": extract_wasDerivedFrom,
     "wasAttributedTo": extract_wasAttributedTo,
     "wasGeneratedBy": extract_session,
@@ -482,14 +600,27 @@ FIELD_EXTRACTORS = {
 }
 
 
-def extract_field(field, metadata):
+def extract_field(field: str, metadata: dict) -> Any:
     if field in FIELD_EXTRACTORS:
         return FIELD_EXTRACTORS[field](metadata)
     else:
-        return metadata.get(field, ...)
+        return metadata.get(field)
 
 
-neurodata_typemap = {
+if TYPE_CHECKING:
+    if sys.version_info >= (3, 8):
+        from typing import TypedDict
+    else:
+        from typing_extensions import TypedDict
+
+    class Neurodatum(TypedDict):
+        module: str
+        neurodata_type: str
+        technique: Optional[str]
+        approach: Optional[str]
+
+
+neurodata_typemap: Dict[str, Neurodatum] = {
     "ElectricalSeries": {
         "module": "ecephys",
         "neurodata_type": "ElectricalSeries",
@@ -679,11 +810,14 @@ neurodata_typemap = {
 }
 
 
-def process_ndtypes(asset, nd_types):
+def process_ndtypes(
+    asset: models.BareAsset, nd_types: Iterable[str]
+) -> models.BareAsset:
     approach = set()
     technique = set()
     variables = set()
     for val in nd_types:
+        val = val.split()[0]
         if val not in neurodata_typemap:
             continue
         if neurodata_typemap[val]["approach"]:
@@ -699,32 +833,10 @@ def process_ndtypes(asset, nd_types):
     return asset
 
 
-def get_asset_metadata(
-    filepath, relpath, digest=None, digest_type=None, allow_any_path=True
-) -> models.BareAsset:
-    metadata = None
-    if op.splitext(filepath)[1] == ".nwb":
-        try:
-            metadata = nwb2asset(filepath, digest=digest, digest_type=digest_type)
-        except Exception as e:
-            lgr.warning(
-                "Failed to extract NWB metadata from %s: %s: %s",
-                filepath,
-                type(e).__name__,
-                str(e),
-            )
-            if not allow_any_path:
-                raise
-    if metadata is None:
-        metadata = get_default_metadata(
-            filepath, digest=digest, digest_type=digest_type
-        )
-    metadata.path = str(relpath)
-    return metadata
-
-
 def nwb2asset(
-    nwb_path, digest=None, digest_type=None, schema_version=None
+    nwb_path: Union[str, Path],
+    digest: Optional[Digest] = None,
+    schema_version: Optional[str] = None,
 ) -> models.BareAsset:
     if schema_version is not None:
         current_version = models.get_schema_version()
@@ -735,13 +847,13 @@ def nwb2asset(
     start_time = datetime.now().astimezone()
     metadata = get_metadata(nwb_path)
     if digest is not None:
-        metadata["digest"] = digest
-        metadata["digest_type"] = digest_type
+        metadata["digest"] = digest.value
+        metadata["digest_type"] = digest.algorithm.name
     metadata["contentSize"] = op.getsize(nwb_path)
     metadata["encodingFormat"] = "application/x-nwb"
     metadata["dateModified"] = get_utcnow_datetime()
     metadata["blobDateModified"] = ensure_datetime(os.stat(nwb_path).st_mtime)
-    metadata["path"] = nwb_path
+    metadata["path"] = str(nwb_path)
     if metadata["blobDateModified"] > metadata["dateModified"]:
         lgr.warning(
             "mtime %s of %s is in the future", metadata["blobDateModified"], nwb_path
@@ -755,12 +867,14 @@ def nwb2asset(
     return asset
 
 
-def get_default_metadata(path, digest=None, digest_type=None) -> models.BareAsset:
+def get_default_metadata(
+    path: Union[str, Path], digest: Optional[Digest] = None
+) -> models.BareAsset:
     start_time = datetime.now().astimezone()
     if digest is not None:
-        digest_model = {models.DigestType[digest_type]: digest}
+        digest_model = digest.asdict()
     else:
-        digest_model = []
+        digest_model = {}
     dateModified = get_utcnow_datetime()
     blobDateModified = ensure_datetime(os.stat(path).st_mtime)
     if blobDateModified > dateModified:
@@ -772,7 +886,7 @@ def get_default_metadata(path, digest=None, digest_type=None) -> models.BareAsse
         dateModified=dateModified,
         blobDateModified=blobDateModified,
         wasGeneratedBy=[get_generator(start_time, end_time)],
-        encodingFormat=get_mime_type(path),
+        encodingFormat=get_mime_type(str(path)),
     )
 
 
@@ -795,6 +909,6 @@ def get_generator(start_time: datetime, end_time: datetime) -> models.Activity:
     )
 
 
-def metadata2asset(metadata):
+def metadata2asset(metadata: dict) -> models.BareAsset:
     bare_dict = extract_model(models.BareAsset, metadata).json_dict()
     return models.BareAsset(**bare_dict)
